//
//  CpabeTest.m
//  dawson
//
//  Created by Neil Spring on 1/19/09.
//  Copyright 2009 University of Maryland. All rights reserved.
//

#import "CpabeTest.h"


@implementation CpabeTest

@end

// open ssl libs
#include <openssl/ssl.h>
#include <openssl/err.h>
#include <openssl/rsa.h>
#include <openssl/engine.h>
#include <openssl/bio.h>
#include <openssl/evp.h>

// ciphertext policy attribute based encryption libraries
#include <glibconfig.h>
#include <glib.h>
#include <pbc/pbc.h>
#include <pbc/pbc_random.h>

#include <errno.h>
#include <bswabe.h>
#include <cpabe/common.h>
#include <cpabe/policy_lang.h>

#include <stdio.h> // for printf() (really?)
#include <sys/time.h>

#define RSA_KEYSIZE 1024
#define BUFSIZE 2048

char *base64_encode(char *str, int len);
int base64_decode(char *in, char **out);
void unserialize_cpabe_buf(char *buf, GByteArray **cph, int *file_len, GByteArray **aes);
GByteArray *suck_buffer(char * buffer, int buf_len);
char *serialize_cpabe_buf(GByteArray *cph, int len, GByteArray *aes, int *buf_len);
char *base64_encode_array(GByteArray *a);
gint comp_string( gconstpointer a, gconstpointer b);
GByteArray *base64_decode_array(const char *encoded);
static int inner_decrypt(bswabe_pub_t *pub, bswabe_prv_t *prv, const char *data, char **plainText);
static void inner_EncryptABE(const char *data, const char *accessstructure,
							 bswabe_pub_t *pub, char ** cipherText);




int OldCpabeTestDo (void) {
    
  char * ampk;
  char * amsk;
  char * sk;
  char * cipherText;
  char * plainText;
  
  SSL_load_error_strings();
  SSL_library_init();
  OpenSSL_add_all_algorithms();
  
  
  GenerateABEKeyPair(&ampk, &amsk);
  
  printf("ampk: %s\n\n", ampk);
  printf("amsk: %s\n\n", amsk);
  
  //bswabe_pub_t *pub = bswabe_pub_unserialize(base64_decode_array(ampk), 1);
  
  GenerateSK("friend family", ampk,
                   amsk, &sk);
  printf("sk: %s\n\n", sk);
  
  //bswabe_prv_t *prv = bswabe_prv_unserialize(pub, base64_decode_array(sk), 1);
	
	NSDate *start = [NSDate	date];
	int i;
	for(i=0; i<20; i++) {
		GenerateABEKeyPair(&ampk, &amsk);
	}
	NSDate *end = [NSDate date];
	NSLog(@"ABE Keypair generation time: %5.3f / 20 = %5.3fs", [end timeIntervalSinceDate:start], [end timeIntervalSinceDate:start]/20);
	
	printf("ampk: %s\n\n", ampk);
	printf("amsk: %s\n\n", amsk);
	
	start = [NSDate	date];
	for(i=0; i<20; i++) {
		GenerateSK("friend family", ampk,
                   amsk, &sk);
	}
	end = [NSDate date];
	NSLog(@"ABE attribute sk generation time: %5.3f / 20 = %5.3fs", [end timeIntervalSinceDate:start], [end timeIntervalSinceDate:start]/20);
	
	printf("sk: %s\n\n", sk);
	
	start = [NSDate	date];
	for(i=0; i<20; i++) {
		EncryptABE("hello world", "friend",
                   ampk, &cipherText);
	}
	end = [NSDate date];
	NSLog(@"ABE encrypt  time: %5.3f / 20 = %5.3fs", [end timeIntervalSinceDate:start], [end timeIntervalSinceDate:start]/20);
	printf("cipherText: %s\n\n", cipherText);
	
	start = [NSDate	date];
	for(i=0; i<20; i++) {
		DecryptABE(cipherText, ampk,
                   sk, &plainText);
		
	}
	end = [NSDate date];
	NSLog(@"ABE decrypt  time: %5.3f / 20 = %5.3fs", [end timeIntervalSinceDate:start], [end timeIntervalSinceDate:start]/20);
	
	printf("plainText: %s\n\n", plainText);
	
	free(ampk);
	free(amsk);
	free(sk);
	free(cipherText);
	free(plainText);
	
	return 0;
	
}	
  

int CpabeTestDo (void) {
    
	char * ampk_b64;
	char * amsk_b64;
	char * sk_b64;
	char * cipherText;
	char * plainText;
	
	SSL_load_error_strings();
	SSL_library_init();
	OpenSSL_add_all_algorithms();
	
	
	GenerateABEKeyPair(&ampk_b64, &amsk_b64);
	
	printf("ampk: %s\n\n", ampk_b64);
	printf("amsk: %s\n\n", amsk_b64);
	
	bswabe_pub_t *pub = bswabe_pub_unserialize(base64_decode_array(ampk_b64), 1);
	
	GenerateSK("friend family", ampk_b64,
			   amsk_b64, &sk_b64);
	printf("sk: %s\n\n", sk_b64);
	
	bswabe_prv_t *prv = bswabe_prv_unserialize(pub, base64_decode_array(sk_b64), 1);
	
	printf("building encrypted messages\n");
	int i;
#define NUM_ATTRS 5
#define NUM_MESSAGES 40
#define NUM_SIZES 10
#define MAX_STRING_SIZE 4096
	
	char *encrypted[NUM_ATTRS][NUM_SIZES][NUM_MESSAGES];
	char access_structure[1024];
	char random_string[MAX_STRING_SIZE];
	int random_string_lengths[NUM_SIZES];
	int sz, attrs;
	struct timeval start, end;
	struct timeval deltat; 
	
	for(sz=0; sz< NUM_SIZES; sz++) {
		random_string_lengths[sz] = rand() % MAX_STRING_SIZE;
	}

	gettimeofday(&start,NULL);
	strcpy(access_structure, "friend");
	for(attrs = 0; attrs < NUM_ATTRS; attrs++) {
		for(sz=0; sz< NUM_SIZES; sz++) {
			for(i=0; i<NUM_MESSAGES; i++) {
				int j;
				for(j=0;j<random_string_lengths[sz];j++) {
					random_string[j] = rand() % 255 + 1; /* not null. */
				}
				random_string[j] = '\0';
				inner_EncryptABE(random_string, access_structure, pub, &encrypted[attrs][sz][i]);
			}
			gettimeofday(&end, NULL);
			
			timersub(&end, &start, &deltat);
			
			printf("finished encrypting random blobs, size %d of %d (%d) after %lu.%06lu using access structure %s\n", sz, NUM_SIZES, random_string_lengths[sz], deltat.tv_sec, deltat.tv_usec, access_structure);
		}
		
		access_structure[0]='\0';
		strcat(access_structure, "bogonXXX");
		access_structure[strlen(access_structure)-1] = 'a' + rand() % 25;
		access_structure[strlen(access_structure)-2] = 'a' + rand() % 25;
		access_structure[strlen(access_structure)-3] = 'a' + rand() % 25;
		
		
		for(i=0; i< attrs; i++) {
			strcat(access_structure, " or bogonXXX");
			access_structure[strlen(access_structure)-1] = 'A' + rand() % 25;
			access_structure[strlen(access_structure)-2] = 'A' + rand() % 25;
			access_structure[strlen(access_structure)-3] = 'A' + rand() % 25;
		
			
		#define BIGASS false
			if (BIGASS) {

			
			strcat(access_structure, " or bogonXXX");
			access_structure[strlen(access_structure)-1] = 'A' + rand() % 25;
			access_structure[strlen(access_structure)-2] = 'A' + rand() % 25;
			access_structure[strlen(access_structure)-3] = 'A' + rand() % 25;
			
			strcat(access_structure, " or bogonXXX");
			access_structure[strlen(access_structure)-1] = 'A' + rand() % 25;
			access_structure[strlen(access_structure)-2] = 'A' + rand() % 25;
			access_structure[strlen(access_structure)-3] = 'A' + rand() % 25;
			
			strcat(access_structure, " or bogonXXX");
			access_structure[strlen(access_structure)-1] = 'A' + rand() % 25;
			access_structure[strlen(access_structure)-2] = 'A' + rand() % 25;
			access_structure[strlen(access_structure)-3] = 'A' + rand() % 25;
			
			strcat(access_structure, " or bogonXXX");
			access_structure[strlen(access_structure)-1] = 'A' + rand() % 25;
			access_structure[strlen(access_structure)-2] = 'A' + rand() % 25;
			access_structure[strlen(access_structure)-3] = 'A' + rand() % 25;
			
			strcat(access_structure, " or bogonXXX");
			access_structure[strlen(access_structure)-1] = 'A' + rand() % 25;
			access_structure[strlen(access_structure)-2] = 'A' + rand() % 25;
			access_structure[strlen(access_structure)-3] = 'A' + rand() % 25;
			
			strcat(access_structure, " or bogonXXX");
			access_structure[strlen(access_structure)-1] = 'A' + rand() % 25;
			access_structure[strlen(access_structure)-2] = 'A' + rand() % 25;
			access_structure[strlen(access_structure)-3] = 'A' + rand() % 25;
			
			strcat(access_structure, " or bogonXXX");
			access_structure[strlen(access_structure)-1] = 'A' + rand() % 25;
			access_structure[strlen(access_structure)-2] = 'A' + rand() % 25;
			access_structure[strlen(access_structure)-3] = 'A' + rand() % 25;
			
			strcat(access_structure, " or bogonXXX");
			access_structure[strlen(access_structure)-1] = 'A' + rand() % 25;
			access_structure[strlen(access_structure)-2] = 'A' + rand() % 25;
			access_structure[strlen(access_structure)-3] = 'A' + rand() % 25;
			
			strcat(access_structure, " or bogonXXX");
			access_structure[strlen(access_structure)-1] = 'A' + rand() % 25;
			access_structure[strlen(access_structure)-2] = 'A' + rand() % 25;
			access_structure[strlen(access_structure)-3] = 'A' + rand() % 25;
			
			strcat(access_structure, " or bogonXXX");
			access_structure[strlen(access_structure)-1] = 'A' + rand() % 25;
			access_structure[strlen(access_structure)-2] = 'A' + rand() % 25;
			access_structure[strlen(access_structure)-3] = 'A' + rand() % 25;
			
			strcat(access_structure, " or bogonXXX");
			access_structure[strlen(access_structure)-1] = 'A' + rand() % 25;
			access_structure[strlen(access_structure)-2] = 'A' + rand() % 25;
			access_structure[strlen(access_structure)-3] = 'A' + rand() % 25;
			
			strcat(access_structure, " or bogonXXX");
			access_structure[strlen(access_structure)-1] = 'A' + rand() % 25;
			access_structure[strlen(access_structure)-2] = 'A' + rand() % 25;
			access_structure[strlen(access_structure)-3] = 'A' + rand() % 25;
			
			strcat(access_structure, " or bogonXXX");
			access_structure[strlen(access_structure)-1] = 'A' + rand() % 25;
			access_structure[strlen(access_structure)-2] = 'A' + rand() % 25;
			access_structure[strlen(access_structure)-3] = 'A' + rand() % 25;
			
			strcat(access_structure, " or bogonXXX");
			access_structure[strlen(access_structure)-1] = 'A' + rand() % 25;
			access_structure[strlen(access_structure)-2] = 'A' + rand() % 25;
			access_structure[strlen(access_structure)-3] = 'A' + rand() % 25;
			
			strcat(access_structure, " or bogonXXX");
			access_structure[strlen(access_structure)-1] = 'A' + rand() % 25;
			access_structure[strlen(access_structure)-2] = 'A' + rand() % 25;
			access_structure[strlen(access_structure)-3] = 'A' + rand() % 25;
			
			strcat(access_structure, " or bogonXXX");
			access_structure[strlen(access_structure)-1] = 'A' + rand() % 25;
			access_structure[strlen(access_structure)-2] = 'A' + rand() % 25;
			access_structure[strlen(access_structure)-3] = 'A' + rand() % 25;
			
			strcat(access_structure, " or bogonXXX");
			access_structure[strlen(access_structure)-1] = 'A' + rand() % 25;
			access_structure[strlen(access_structure)-2] = 'A' + rand() % 25;
			access_structure[strlen(access_structure)-3] = 'A' + rand() % 25;
			
			strcat(access_structure, " or bogonXXX");
			access_structure[strlen(access_structure)-1] = 'A' + rand() % 25;
			access_structure[strlen(access_structure)-2] = 'A' + rand() % 25;
			access_structure[strlen(access_structure)-3] = 'A' + rand() % 25;
			
			strcat(access_structure, " or bogonXXX");
			access_structure[strlen(access_structure)-1] = 'A' + rand() % 25;
			access_structure[strlen(access_structure)-2] = 'A' + rand() % 25;
			access_structure[strlen(access_structure)-3] = 'A' + rand() % 25;
			
			strcat(access_structure, " or bogonXXX");
			access_structure[strlen(access_structure)-1] = 'A' + rand() % 25;
			access_structure[strlen(access_structure)-2] = 'A' + rand() % 25;
			access_structure[strlen(access_structure)-3] = 'A' + rand() % 25;
			} /* bigass */
		}
		
			strcat(access_structure, " or friend");
		
		printf("setting access structure to: '%s'", access_structure);
		
	}
	

	printf("begin decryption test\n");
	
	for(attrs = 0; attrs < NUM_ATTRS; attrs++) {

	for(sz=0; sz< NUM_SIZES; sz++) {
		
		gettimeofday(&start, NULL);
		
		char *dumpbuffer;
		for(i=0; i<NUM_MESSAGES; i++) {
			inner_decrypt(pub, prv, encrypted[attrs][sz][i], &dumpbuffer);
			free(dumpbuffer);
		}
		
		
		gettimeofday(&end, NULL);
		
		timersub(&end, &start, &deltat);
		printf("delta t for %d elements of size %d attrs %d is %lu.%06lu\n", NUM_MESSAGES, random_string_lengths[sz], ( attrs + 1 ) * 2, deltat.tv_sec, deltat.tv_usec);
	}
	
	}
	printf("end decryption test\n");
	
	
	free(ampk_b64);
	free(amsk_b64);
	free(sk_b64);
	//free(cipherText);
	//free(plainText);
	
	return 0;
	
}

/*****************************************************************************/
/*** ABE Functions ***********************************************************/
/*****************************************************************************/

// Function returns a string of the format "pubkey-privkey".  Yes that's
// a dash in there.
void GenerateABEKeyPair(char ** ampk, char ** amsk)
{
  printf("Begin GenerateABEKeyPair\n");
  // generate public and master key from cp-abe
  bswabe_pub_t* pub;
  bswabe_msk_t* msk;
  
  bswabe_setup(&pub, &msk);
  
  char *pub_encoded = base64_encode_array(bswabe_pub_serialize(pub));
  char *msk_encoded = base64_encode_array(bswabe_msk_serialize(msk));
  
  *ampk = (char *) malloc (strlen(pub_encoded)+1);
  *amsk = (char *) malloc (strlen(msk_encoded)+1);
  
  memcpy(*ampk, pub_encoded, strlen(pub_encoded)+1);
  memcpy(*amsk, msk_encoded, strlen(msk_encoded)+1);
  
  free(pub_encoded);
  free(msk_encoded);
  
  printf("End GenerateABEKeyPair\n");
  return;
}

void GenerateSK(const char *attributes, const char *ampk,
                       const char *amsk, char ** sk)
{
	printf("Begin GenerateSK\n");
  // using attributes and master key,
  // generate private key, convert to base 64,
  // and return to privKeyText
  
  bswabe_pub_t* pub;
  bswabe_msk_t* msk;
  bswabe_prv_t* prv;
  char** attrs;
  int i;
  int n;
  GSList* alist;
  GSList* ap;
  char * attrBuf;
  char * attr;
  char * tmpAttr;
  
  pub = bswabe_pub_unserialize(base64_decode_array(ampk), 1);
  msk = bswabe_msk_unserialize(pub, base64_decode_array(amsk), 1);
  
  // process attributes
  alist = 0;
  i = 0;
  int atLength = strlen(attributes);
  
  attrBuf = (char *) malloc (atLength + 1);
  memcpy(attrBuf, attributes, atLength);
  attrBuf[atLength] = '\0';
  
  // now use string tokenizer
  attr = strtok(attrBuf, " ");
  
  while (attr != NULL) {
    tmpAttr = (char *) malloc (strlen(attr) + 1);
    memcpy(tmpAttr, attr, strlen(attr) + 1);
    parse_attribute(&alist, tmpAttr);
    attr = strtok(NULL, " ");
  }
  
  free(attrBuf);
  
  alist = g_slist_sort(alist, comp_string);
  n = g_slist_length(alist);
  
  attrs = (char **)malloc((n + 1) * sizeof(char*));
  
  i = 0;
  for( ap = alist; ap; ap = ap->next )
    attrs[i++] = (char *)ap->data;
  attrs[i] = 0;
  
  // generate private key
  prv = bswabe_keygen(pub, msk, attrs);
  char* prv_encoded = base64_encode_array(bswabe_prv_serialize(prv));
  
  *sk = (char *) malloc (strlen(prv_encoded)+1);
  
  memcpy(*sk, prv_encoded, strlen(prv_encoded)+1);
  
  free(attrs);
  free(prv_encoded);
  
  printf("End GenerateSK\n");
  return;
}


void EncryptABE(const char *data, const char *accessstructure,
				const char *ampk, char ** cipherText) {
	bswabe_pub_t *pub = bswabe_pub_unserialize(base64_decode_array(ampk), 1);
	inner_EncryptABE(data, accessstructure, pub, cipherText);
}
	

static void inner_EncryptABE(const char *data, const char *accessstructure,
				bswabe_pub_t *pub, char ** cipherText)
{
  
  bswabe_cph_t* cph;
  int plt_len;
  GByteArray* plt;
  GByteArray* cph_buf;
  GByteArray* aes_buf;
  element_t m;
  char * policy;
  char * as = (char *)accessstructure;
  char * pt = (char *)data;
  
  // get public key from ampk (which is base64 encoded)
  //pub = bswabe_pub_unserialize(base64_decode_array(ampk), 1);
  
  policy = parse_policy_lang(as);
  cph = bswabe_enc(pub, m, policy);
  free(policy);
  
  cph_buf = bswabe_cph_serialize(cph);
  bswabe_cph_free(cph);
  
  int dtLength = strlen(data);
  
  plt = suck_buffer(pt, (int)dtLength);
  
  plt_len = plt->len;
  
  aes_buf = aes_128_cbc_encrypt(plt, m);
  g_byte_array_free(plt, 1);
  element_clear(m);
  
  int cpabe_len;
  char *cpabe_buf = serialize_cpabe_buf(cph_buf, plt_len, aes_buf, &cpabe_len);
  
  g_byte_array_free(cph_buf, 1);
  g_byte_array_free(aes_buf, 1);
  
  char *encoded = base64_encode(cpabe_buf, cpabe_len);
  
  *cipherText = (char *) malloc (strlen(encoded)+1);
  memcpy(*cipherText, encoded, strlen(encoded)+1);
  
  free(encoded);
  free(cpabe_buf);
  
//  printf("End EncryptABE\n");
  return;
}

  
static int inner_decrypt(bswabe_pub_t *pub, bswabe_prv_t *prv, const char *data, char **plainText) {
    int file_len;
    GByteArray* aes_buf;
    GByteArray* plt;
    GByteArray* cph_buf;
    bswabe_cph_t* cph;
    element_t m;
    char * ctBuffer;
    
  base64_decode((char *) data, &ctBuffer);  // already contains length info
  unserialize_cpabe_buf(ctBuffer, &cph_buf, &file_len, &aes_buf);
  
  cph = bswabe_cph_unserialize(pub, cph_buf, 1);
  if( !bswabe_dec(pub, prv, cph, m) ){
    printf("End DecryptABE\n");
    return 0;
  }
  
  bswabe_cph_free(cph);
  
  plt = aes_128_cbc_decrypt(aes_buf, m);
  
  // DAN, what is this line for?
  g_byte_array_set_size(plt, file_len);
  
  *plainText = (char *) malloc ((int)plt->len+1);
  memcpy(*plainText, (char*) plt->data, (int)plt->len+1);
  
  g_byte_array_free(aes_buf, 1);
  free(ctBuffer);
  
  
  return 1;
  
}

int DecryptABE(const char *data, const char *ampk,
               const char *sk, char ** plainText)
{
	printf("Begin DecryptABE\n");
  
  bswabe_pub_t *pub = bswabe_pub_unserialize(base64_decode_array(ampk), 1);
  bswabe_prv_t *prv = bswabe_prv_unserialize(pub, base64_decode_array(sk), 1);
  
  inner_decrypt(pub, prv, data, plainText);
  
  free(pub);
  free(prv);
  
  printf("End DecryptABE\n");
	
	return 1;
  
}


/***********************************************************************/
/******** begin helper functions ***************************************/
/***********************************************************************/

// returns a null-terminated ascii string
char *base64_encode(char *str, int len) {
  //  printf("Begin base64_encode\n");
  char *cp;
  BIO *bio, *b64;
  int size = (len*4 + 2) / 3;  // account for 4/3 expansion
  size = ((size/4) + 1) * 4 + 1;  // make divible by 4 and add room for '\0'
  char *buf = (char *) malloc(size);
  
  b64 = BIO_new(BIO_f_base64());
  bio = BIO_new(BIO_s_mem());
  BIO_set_flags(b64, BIO_FLAGS_BASE64_NO_NL);
  bio = BIO_push(b64, bio);
  BIO_write(bio, str, len);
  BIO_flush(bio);
  
  int eol = BIO_get_mem_data(bio, &cp);
  memcpy(buf, cp, eol);
  buf[eol] = '\0';
  
  BIO_free_all(bio);
  
  //  printf("End base64_encode\n");
  return buf;
}

int base64_decode(char *in, char **out) {
  //  printf("Begin base64_decode\n");
  BIO *b64, *bmem, *bio;
  char *buf = (char *) malloc(strlen(in));  // output shrinks
  
  b64 = BIO_new(BIO_f_base64());
  BIO_set_flags(b64, BIO_FLAGS_BASE64_NO_NL);
  bmem = BIO_new(BIO_s_mem());
  bio = BIO_push(b64, bmem);
  BIO_write(bmem, in, strlen(in));
  
  int ret = BIO_read(bio, buf, strlen(in));
  buf[ret] = '\0';
  *out = buf;
  
  BIO_free_all(bio);
  
  //  printf("End base64_decode\n");
  return ret;
}

GByteArray *base64_decode_array(const char *encoded) {
  char *result;
  int buf_len = base64_decode((char *) encoded, &result);
  GByteArray *a = suck_buffer(result, buf_len);
  free(result);
  return a;
}

char *base64_encode_array(GByteArray *a) {
  return base64_encode((char *) (a->data), a->len);
}

gint comp_string( gconstpointer a, gconstpointer b) {
	return strcmp(  (const char *)a,   (const char *)b);
}

GByteArray *suck_buffer(char * buffer, int buf_len) {
  GByteArray *a = g_byte_array_sized_new(buf_len);
  return g_byte_array_append(a, (guint8 *) buffer, buf_len);
}

char *serialize_cpabe_buf(GByteArray *cph, int len, GByteArray *aes, int *buf_len) {
  *buf_len = 3*sizeof(int) + cph->len + aes->len;
  
  char *buf = (char *) malloc(*buf_len);
  int index = 0, i;
  
  for (i = 3; i >= 0; i--) {
    buf[index++] = (len & 0xff<<(i*8))>>(i*8);
  }
  for (i = 3; i >= 0; i--) {
    buf[index++] = (aes->len & 0xff<<(i*8))>>(i*8);
  }
  memcpy(buf + index, aes->data, aes->len);
  index += aes->len;
  
  for (i = 3; i >= 0; i--) {
    buf[index++] = (cph->len & 0xff<<(i*8))>>(i*8);
  }
  memcpy(buf + index, cph->data, cph->len);
  index += cph->len;
  
  if (index != *buf_len) {
    printf("error: wrote %d bytes, expected to write %d\n", index, *buf_len);
  }
  
  return buf;
}

void unserialize_cpabe_buf(char *buf, GByteArray **cph, int *file_len, GByteArray **aes) {
  int i, index = 0, len = 0;
  
  *file_len = 0;
  for (i = 3; i >= 0; i--) {
    *file_len |= (buf[index++] & 0xFF) << (i * 8);
  }
  
	for (i = 3; i >= 0; i-- ) {
		len |= (buf[index++] & 0xFF) << (i * 8);
  }
  *aes = g_byte_array_sized_new(len);
  g_byte_array_append(*aes, (guint8 *) buf + index, len);
  index += len;
  
	len = 0;
	for (i = 3; i >= 0; i--) {
		len |= (buf[index++] & 0xFF) << (i*8);
  }
  *cph = g_byte_array_sized_new(len);
  g_byte_array_append(*cph, (guint8 *) buf + index, len);
  //  index += len;  // useless
  
  return;
}
